import numpy as np
import scipy
import centrosome
import centrosome.cpmorphology
import centrosome.filter
import centrosome.propagate
import centrosome.fastemd
from sklearn.cluster import KMeans

from cellprofiler_library.opts import measureimageoverlap as mio
from cellprofiler_library.functions.segmentation import convert_labels_to_ijv
from cellprofiler_library.functions.segmentation import indices_from_ijv
from cellprofiler_library.functions.segmentation import count_from_ijv
from cellprofiler_library.functions.segmentation import areas_from_ijv
from cellprofiler_library.functions.segmentation import cast_labels_to_label_set


def measure_image_overlap_statistics(
    ground_truth_image,
    test_image,
    mask=None,
):
    # Check that the inputs are binary
    if not np.array_equal(ground_truth_image, ground_truth_image.astype(bool)):
        raise ValueError("ground_truth_image is not a binary image")
    
    if not np.array_equal(test_image, test_image.astype(bool)):
        raise ValueError("test_image is not a binary image")

    if mask is None:
        mask = np.ones_like(ground_truth_image, bool)

    orig_shape = ground_truth_image.shape
    
    # Covert 3D image to 2D long
    if ground_truth_image.ndim > 2:
        
        ground_truth_image = ground_truth_image.reshape(
            -1, ground_truth_image.shape[-1]
        )
        test_image = test_image.reshape(-1, test_image.shape[-1])

        mask = mask.reshape(-1, mask.shape[-1])

    false_positives = test_image & ~ground_truth_image

    false_positives[~mask] = False

    false_negatives = (~test_image) & ground_truth_image

    false_negatives[~mask] = False

    true_positives = test_image & ground_truth_image

    true_positives[~mask] = False

    true_negatives = (~test_image) & (~ground_truth_image)

    true_negatives[~mask] = False

    false_positive_count = np.sum(false_positives)

    true_positive_count = np.sum(true_positives)

    false_negative_count = np.sum(false_negatives)

    true_negative_count = np.sum(true_negatives)

    labeled_pixel_count = true_positive_count + false_positive_count

    true_count = true_positive_count + false_negative_count

    if labeled_pixel_count == 0:
        precision = 1.0
    else:
        precision = float(true_positive_count) / float(labeled_pixel_count)

    if true_count == 0:
        recall = 1.0
    else:
        recall = float(true_positive_count) / float(true_count)

    if (precision + recall) == 0:
        f_factor = 0.0  # From http://en.wikipedia.org/wiki/F1_score
    else:
        f_factor = 2.0 * precision * recall / (precision + recall)

    negative_count = false_positive_count + true_negative_count

    if negative_count == 0:
        false_positive_rate = 0.0

        true_negative_rate = 1.0
    else:
        false_positive_rate = float(false_positive_count) / float(negative_count)

        true_negative_rate = float(true_negative_count) / float(negative_count)
    if true_count == 0:
        false_negative_rate = 0.0

        true_positive_rate = 1.0
    else:
        false_negative_rate = float(false_negative_count) / float(true_count)

        true_positive_rate = float(true_positive_count) / float(true_count)

    ground_truth_labels, ground_truth_count = scipy.ndimage.label(
        ground_truth_image & mask, np.ones((3, 3), bool)
    )

    test_labels, test_count = scipy.ndimage.label(
        test_image & mask, np.ones((3, 3), bool)
    )

    rand_index, adjusted_rand_index = compute_rand_index(
        test_labels, ground_truth_labels, mask
    )

    data = {
        "true_positives": true_positives.reshape(orig_shape),
        "true_negatives": true_negatives.reshape(orig_shape),
        "false_positives": false_positives.reshape(orig_shape),
        "false_negatives": false_negatives.reshape(orig_shape),
        "Ffactor": f_factor,
        "Precision": precision,
        "Recall": recall,
        "TruePosRate": true_positive_rate,
        "FalsePosRate": false_positive_rate,
        "FalseNegRate": false_negative_rate,
        "TrueNegRate": true_negative_rate,
        "RandIndex": rand_index,
        "AdjustedRandIndex": adjusted_rand_index,
    }

    return data


def compute_rand_index(test_labels, ground_truth_labels, mask):
    """Calculate the Rand Index

    http://en.wikipedia.org/wiki/Rand_index

    Given a set of N elements and two partitions of that set, X and Y

    A = the number of pairs of elements in S that are in the same set in
        X and in the same set in Y
    B = the number of pairs of elements in S that are in different sets
        in X and different sets in Y
    C = the number of pairs of elements in S that are in the same set in
        X and different sets in Y
    D = the number of pairs of elements in S that are in different sets
        in X and the same set in Y

    The rand index is:   A + B
                            -----
                        A+B+C+D


    The adjusted rand index is the rand index adjusted for chance
    so as not to penalize situations with many segmentations.

    Jorge M. Santos, Mark Embrechts, "On the Use of the Adjusted Rand
    Index as a Metric for Evaluating Supervised Classification",
    Lecture Notes in Computer Science,
    Springer, Vol. 5769, pp. 175-184, 2009. Eqn # 6

    ExpectedIndex = best possible score

    ExpectedIndex = sum(N_i choose 2) * sum(N_j choose 2)

    MaxIndex = worst possible score = 1/2 (sum(N_i choose 2) + sum(N_j choose 2)) * total

    A * total - ExpectedIndex
    -------------------------
    MaxIndex - ExpectedIndex

    returns a tuple of the Rand Index and the adjusted Rand Index
    """
    ground_truth_labels = ground_truth_labels[mask].astype(np.uint32)
    test_labels = test_labels[mask].astype(np.uint32)
    if len(test_labels) > 0:
        #
        # Create a sparse matrix of the pixel labels in each of the sets
        #
        # The matrix, N(i,j) gives the counts of all of the pixels that were
        # labeled with label I in the ground truth and label J in the
        # test set.
        #
        N_ij = scipy.sparse.coo_matrix(
            (np.ones(len(test_labels)), (ground_truth_labels, test_labels))
        ).toarray()

        def choose2(x):
            """Compute # of pairs of x things = x * (x-1) / 2"""
            return x * (x - 1) / 2

        #
        # Each cell in the matrix is a count of a grouping of pixels whose
        # pixel pairs are in the same set in both groups. The number of
        # pixel pairs is n * (n - 1), so A = sum(matrix * (matrix - 1))
        #
        A = np.sum(choose2(N_ij))
        #
        # B is the sum of pixels that were classified differently by both
        # sets. But the easier calculation is to find A, C and D and get
        # B by subtracting A, C and D from the N * (N - 1), the total
        # number of pairs.
        #
        # For C, we take the number of pixels classified as "i" and for each
        # "j", subtract N(i,j) from N(i) to get the number of pixels in
        # N(i,j) that are in some other set = (N(i) - N(i,j)) * N(i,j)
        #
        # We do the similar calculation for D
        #
        N_i = np.sum(N_ij, 1)
        N_j = np.sum(N_ij, 0)
        C = np.sum((N_i[:, np.newaxis] - N_ij) * N_ij) / 2
        D = np.sum((N_j[np.newaxis, :] - N_ij) * N_ij) / 2
        total = choose2(len(test_labels))
        # an astute observer would say, why bother computing A and B
        # when all we need is A+B and C, D and the total can be used to do
        # that. The calculations aren't too expensive, though, so I do them.
        B = total - A - C - D
        rand_index = (A + B) / total
        #
        # Compute adjusted Rand Index
        #
        expected_index = np.sum(choose2(N_i)) * np.sum(choose2(N_j))
        max_index = (np.sum(choose2(N_i)) + np.sum(choose2(N_j))) * total / 2

        adjusted_rand_index = (A * total - expected_index) / (
            max_index - expected_index
        )
    else:
        rand_index = adjusted_rand_index = np.nan
    return rand_index, adjusted_rand_index


def compute_earth_movers_distance(
    ground_truth_image,
    test_image,
    mask=None,
    decimation_method: mio.DM = mio.DM.KMEANS,
    max_distance: int = 250,
    max_points: int = 250,
    penalize_missing: bool = False,
):
    """Compute the earthmovers distance between two sets of objects

    src_objects - move pixels from these objects

    dest_objects - move pixels to these objects

    returns the earth mover's distance
    """

    # Check that the inputs are binary
    if not np.array_equal(ground_truth_image, ground_truth_image.astype(bool)):
        raise ValueError("ground_truth_image is not a binary image")
    
    if not np.array_equal(test_image, test_image.astype(bool)):
        raise ValueError("test_image is not a binary image")

    if mask is None:
        mask = np.ones_like(ground_truth_image, bool)

    # Covert 3D image to 2D long
    if ground_truth_image.ndim > 2:
        ground_truth_image = ground_truth_image.reshape(
            -1, ground_truth_image.shape[-1]
        )

        test_image = test_image.reshape(-1, test_image.shape[-1])

        mask = mask.reshape(-1, mask.shape[-1])

    # ground truth labels
    dest_labels = scipy.ndimage.label(
        ground_truth_image & mask, np.ones((3, 3), bool)
    )[0]
    dest_labelset = cast_labels_to_label_set(dest_labels)
    dest_ijv = convert_labels_to_ijv(dest_labels, validate=False)
    dest_ijv_indices = indices_from_ijv(dest_ijv, validate=False)
    dest_count = count_from_ijv(
        dest_ijv, indices=dest_ijv_indices, validate=False)
    dest_areas = areas_from_ijv(
        dest_ijv, indices=dest_ijv_indices, validate=False)

    # test labels
    src_labels = scipy.ndimage.label(
        test_image & mask, np.ones((3, 3), bool)
    )[0]
    src_labelset = cast_labels_to_label_set(src_labels)
    src_ijv = convert_labels_to_ijv(src_labels, validate=False)
    src_ijv_indices = indices_from_ijv(src_ijv, validate=False)
    src_count = count_from_ijv(
        src_ijv, indices=src_ijv_indices, validate=False)
    src_areas = areas_from_ijv(
        src_ijv, indices=src_ijv_indices, validate=False)

    #
    # if either foreground set is empty, the emd is the penalty.
    #
    for lef_count, right_areas in (
        (src_count, dest_areas),
        (dest_count, src_areas),
    ):
        if lef_count == 0:
            if penalize_missing:
                return np.sum(right_areas) * max_distance
            else:
                return 0
    if decimation_method == mio.DM.KMEANS:
        isrc, jsrc = get_kmeans_points(src_ijv, dest_ijv, max_points)
        idest, jdest = isrc, jsrc
    elif decimation_method == mio.DM.SKELETON:
        isrc, jsrc = get_skeleton_points(src_labelset, src_labels.shape, max_points)
        idest, jdest = get_skeleton_points(dest_labelset, dest_labels.shape, max_points)
    else:
        raise TypeError("Unknown type for decimation method: %s" % decimation_method)
    src_weights, dest_weights = [
        get_weights(i, j, get_labels_mask(labelset, shape))
        for i, j, labelset, shape in (
            (isrc, jsrc, src_labelset, src_labels.shape),
            (idest, jdest, dest_labelset, dest_labels.shape),
        )
    ]
    ioff, joff = [
        src[:, np.newaxis] - dest[np.newaxis, :]
        for src, dest in ((isrc, idest), (jsrc, jdest))
    ]
    c = np.sqrt(ioff * ioff + joff * joff).astype(np.int32)
    c[c > max_distance] = max_distance
    extra_mass_penalty = max_distance if penalize_missing else 0

    emd = centrosome.fastemd.emd_hat_int32(
        src_weights.astype(np.int32),
        dest_weights.astype(np.int32),
        c,
        extra_mass_penalty=extra_mass_penalty,
    )
    return emd


def get_labels_mask(labelset, shape):
    labels_mask = np.zeros(shape, bool)
    for labels, indexes in labelset:
        labels_mask = labels_mask | labels > 0
    return labels_mask


def get_skeleton_points(labelset, shape, max_points):
    """Get points by skeletonizing the objects and decimating"""
    total_skel = np.zeros(shape, bool)

    for labels, indexes in labelset:
        colors = centrosome.cpmorphology.color_labels(labels)
        for color in range(1, np.max(colors) + 1):
            labels_mask = colors == color
            skel = centrosome.cpmorphology.skeletonize(
                labels_mask,
                ordering=scipy.ndimage.distance_transform_edt(labels_mask)
                * centrosome.filter.poisson_equation(labels_mask),
            )
            total_skel = total_skel | skel

    n_pts = np.sum(total_skel)

    if n_pts == 0:
        return np.zeros(0, np.int32), np.zeros(0, np.int32)

    i, j = np.where(total_skel)

    if n_pts > max_points:
        #
        # Decimate the skeleton by finding the branchpoints in the
        # skeleton and propagating from those.
        #
        markers = np.zeros(total_skel.shape, np.int32)
        branchpoints = centrosome.cpmorphology.branchpoints(
            total_skel
        ) | centrosome.cpmorphology.endpoints(total_skel)
        markers[branchpoints] = np.arange(np.sum(branchpoints)) + 1
        #
        # We compute the propagation distance to that point, then impose
        # a slightly arbitrary order to get an unambiguous ordering
        # which should number the pixels in a skeleton branch monotonically
        #
        ts_labels, distances = centrosome.propagate.propagate(
            np.zeros(markers.shape), markers, total_skel, 1
        )
        order = np.lexsort((j, i, distances[i, j], ts_labels[i, j]))
        #
        # Get a linear space of self.max_points elements with bounds at
        # 0 and len(order)-1 and use that to select the points.
        #
        order = order[np.linspace(0, len(order) - 1, max_points).astype(int)]
        return i[order], j[order]

    return i, j


def get_kmeans_points(src_ijv, dest_ijv, max_points):
    """Get representative points in the objects using K means

    src_ijv - get some of the foreground points from the source ijv labeling
    dest_ijv - get the rest of the foreground points from the ijv labeling
                objects

    returns a vector of i coordinates of representatives and a vector
            of j coordinates
    """

    ijv = np.vstack((src_ijv, dest_ijv))
    if len(ijv) <= max_points:
        return ijv[:, 0], ijv[:, 1]
    random_state = np.random.RandomState()
    random_state.seed(ijv.astype(int).flatten())
    kmeans = KMeans(n_clusters=max_points, tol=2, random_state=random_state)
    kmeans.fit(ijv[:, :2])
    return (
        kmeans.cluster_centers_[:, 0].astype(np.uint32),
        kmeans.cluster_centers_[:, 1].astype(np.uint32),
    )


def get_weights(i, j, labels_mask):
    """Return the weights to assign each i,j point

    Assign each pixel in the labels mask to the nearest i,j and return
    the number of pixels assigned to each i,j
    """
    #
    # Create a mapping of chosen points to their index in the i,j array
    #
    total_skel = np.zeros(labels_mask.shape, int)
    total_skel[i, j] = np.arange(1, len(i) + 1)
    #
    # Compute the distance from each chosen point to all others in image,
    # return the nearest point.
    #
    ii, jj = scipy.ndimage.distance_transform_edt(
        total_skel == 0, return_indices=True, return_distances=False
    )
    #
    # Filter out all unmasked points
    #
    ii, jj = [x[labels_mask] for x in (ii, jj)]
    if len(ii) == 0:
        return np.zeros(0, np.int32)
    #
    # Use total_skel to look up the indices of the chosen points and
    # bincount the indices.
    #
    result = np.zeros(len(i), np.int32)
    bc = np.bincount(total_skel[ii, jj])[1:]
    result[: len(bc)] = bc
    return result



from ..functions.object_processing import fill_object_holes, fill_convex_hulls

def fillobjects(labels, mode="holes", diameter=64.0, planewise=False):
    if mode.casefold() == "holes":
        return fill_object_holes(labels, diameter, planewise)
    elif mode.casefold() in ("convex hull", "convex_hull"):
        return fill_convex_hulls(labels)
    else:
        raise ValueError(f"Mode '{mode}' is not supported. Available modes are: 'holes' and 'convex_hull'.")




import numpy

from ..functions.image_processing import rgb_to_greyscale, medial_axis

def medialaxis(image, multichannel, volumetric):
    if multichannel:
        image = rgb_to_greyscale(image)

    if volumetric:
        data = numpy.zeros_like(image)

        for z, plane in enumerate(image):
            data[z] = medial_axis(plane)
        return data
    else:
        return medial_axis(image)


import numpy
import skimage.color
import skimage.morphology
import centrosome
import centrosome.threshold
import scipy
import matplotlib


def rgb_to_greyscale(image):
    if image.shape[-1] == 4:
        output = skimage.color.rgba2rgb(image)
        return skimage.color.rgb2gray(output)
    else:
        return skimage.color.rgb2gray(image)


def medial_axis(image):
    if image.ndim > 2 and image.shape[-1] in (3, 4):
        raise ValueError("Convert image to grayscale or use medialaxis module")
    if image.ndim > 2 and image.shape[-1] not in (3, 4):
        raise ValueError("Process 3D images plane-wise or use the medialaxis module")
    return skimage.morphology.medial_axis(image)


def enhance_edges_sobel(image, mask=None, direction="all"):
    if direction.casefold() == "all":
        output_pixels = centrosome.filter.sobel(image, mask)
    elif direction.casefold() == "horizontal":
        output_pixels = centrosome.filter.hsobel(image, mask)
    elif direction.casefold() == "vertical":
        output_pixels = centrosome.filter.vsobel(image, mask)
    else:
        raise NotImplementedError(f"Unimplemented direction for Sobel: {direction}")
    return output_pixels


def enhance_edges_log(image, mask=None, sigma=2.0):
    size = int(sigma * 4) + 1
    output_pixels = centrosome.filter.laplacian_of_gaussian(image, mask, size, sigma)
    return output_pixels


def enhance_edges_prewitt(image, mask=None, direction="all"):
    if direction.casefold() == "all":
        output_pixels = centrosome.filter.prewitt(image, mask)
    elif direction.casefold() == "horizontal":
        output_pixels = centrosome.filter.hprewitt(image, mask)
    elif direction.casefold() == "vertical":
        output_pixels = centrosome.filter.vprewitt(image, mask)
    else:
        raise NotImplementedError(f"Unimplemented direction for Prewitt: {direction}")
    return output_pixels


def enhance_edges_canny(
    image,
    mask=None,
    auto_threshold=True,
    auto_low_threshold=True,
    sigma=1.0,
    low_threshold=0.1,
    manual_threshold=0.2,
    threshold_adjustment_factor=1.0,
):

    if auto_threshold or auto_low_threshold:
        sobel_image = centrosome.filter.sobel(image)
        low, high = centrosome.otsu.otsu3(sobel_image[mask])
        if auto_threshold:
            high_th = high * threshold_adjustment_factor
        if auto_low_threshold:
            low_th = low * threshold_adjustment_factor
    else:
        low_th = low_threshold
        high_th = manual_threshold

    output_pixels = centrosome.filter.canny(image, mask, sigma, low_th, high_th)
    return output_pixels


def morphology_closing(image, structuring_element=skimage.morphology.disk(1)):
    if structuring_element.ndim == 3 and image.ndim == 2:
        raise ValueError("Cannot apply a 3D structuring element to a 2D image")
    # Check if a 2D structuring element will be applied to a 3D image planewise
    planewise = structuring_element.ndim == 2 and image.ndim == 3
    if planewise:
        output = numpy.zeros_like(image)
        for index, plane in enumerate(image):
            output[index] = skimage.morphology.closing(plane, structuring_element)
        return output
    else:
        return skimage.morphology.closing(image, structuring_element)


def morphology_opening(image, structuring_element=skimage.morphology.disk(1)):
    if structuring_element.ndim == 3 and image.ndim == 2:
        raise ValueError("Cannot apply a 3D structuring element to a 2D image")
    # Check if a 2D structuring element will be applied to a 3D image planewise
    planewise = structuring_element.ndim == 2 and image.ndim == 3
    if planewise:
        output = numpy.zeros_like(image)
        for index, plane in enumerate(image):
            output[index] = skimage.morphology.opening(plane, structuring_element)
        return output
    else:
        return skimage.morphology.opening(image, structuring_element)


def morphological_skeleton_2d(image):
    return skimage.morphology.skeletonize(image)


def morphological_skeleton_3d(image):
    return skimage.morphology.skeletonize_3d(image)


def median_filter(image, window_size, mode):
    return scipy.ndimage.median_filter(image, size=window_size, mode=mode)


def reduce_noise(image, patch_size, patch_distance, cutoff_distance, channel_axis=None):
    denoised = skimage.restoration.denoise_nl_means(
        image=image,
        patch_size=patch_size,
        patch_distance=patch_distance,
        h=cutoff_distance,
        channel_axis=channel_axis,
        fast_mode=True,
    )
    return denoised


def get_threshold_robust_background(
    image,
    lower_outlier_fraction=0.05,
    upper_outlier_fraction=0.05,
    averaging_method="mean",
    variance_method="standard_deviation",
    number_of_deviations=2,
):
    """Calculate threshold based on mean & standard deviation.
    The threshold is calculated by trimming the top and bottom 5% of
    pixels off the image, then calculating the mean and standard deviation
    of the remaining image. The threshold is then set at 2 (empirical
    value) standard deviations above the mean.


    lower_outlier_fraction - after ordering the pixels by intensity, remove
        the pixels from 0 to len(image) * lower_outlier_fraction from
        the threshold calculation (default = 0.05).
    upper_outlier_fraction - remove the pixels from
        len(image) * (1 - upper_outlier_fraction) to len(image) from
        consideration (default = 0.05).
    averaging_method - Determines how the intensity midpoint is determined
        after discarding outliers. (default "Mean". Options: "Mean", "Median",
        "Mode").
    variance_method - Method to calculate variance (default =
        "Standard deviation". Options: "Standard deviation",
        "Median absolute deviation")
    number_of_deviations - Following calculation of the standard deviation
        or MAD, multiply this number and add to the average to get the final
        threshold (default = 2)
    average_fn - function used to calculate the average intensity (e.g.
        np.mean, np.median or some sort of mode function). Default = np.mean
    variance_fn - function used to calculate the amount of variance.
                    Default = np.sd
    """

    if averaging_method.casefold() == "mean":
        average_fn = numpy.mean
    elif averaging_method.casefold() == "median":
        average_fn = numpy.median
    elif averaging_method.casefold() == "mode":
        average_fn = centrosome.threshold.binned_mode
    else:
        raise ValueError(f"{averaging_method} not in 'Mean', 'Median', 'Mode'")

    if variance_method.casefold() == "standard_deviation":
        variance_fn = numpy.std
    elif variance_method.casefold() == "median_absolute_deviation":
        variance_fn = centrosome.threshold.mad
    else:
        raise ValueError(
            f"{variance_method} not in 'standard_deviation', 'median_absolute_deviation'"
        )

    flat_image = image.flatten()
    n_pixels = len(flat_image)
    if n_pixels < 3:
        return 0

    flat_image.sort()
    if flat_image[0] == flat_image[-1]:
        return flat_image[0]
    low_chop = int(round(n_pixels * lower_outlier_fraction))
    hi_chop = n_pixels - int(round(n_pixels * upper_outlier_fraction))
    im = flat_image if low_chop == 0 else flat_image[low_chop:hi_chop]
    mean = average_fn(im)
    sd = variance_fn(im)
    return mean + sd * number_of_deviations


def get_adaptive_threshold(
    image,
    mask=None,
    threshold_method="otsu",
    window_size=50,
    threshold_min=0,
    threshold_max=1,
    threshold_correction_factor=1,
    assign_middle_to_foreground="foreground",
    global_limits=[0.7, 1.5],
    log_transform=False,
    volumetric=False,
    **kwargs,
):

    if mask is not None:
        # Apply mask and preserve image shape
        image = numpy.where(mask, image, False)

    if volumetric:
        # Array to store threshold values
        thresh_out = numpy.zeros(image.shape)
        for z in range(image.shape[0]):
            thresh_out[z, :, :] = get_adaptive_threshold(
                image[z, :, :],
                mask=None,  # Mask has already been applied
                threshold_method=threshold_method,
                window_size=window_size,
                threshold_min=threshold_min,
                threshold_max=threshold_max,
                threshold_correction_factor=threshold_correction_factor,
                assign_middle_to_foreground=assign_middle_to_foreground,
                global_limits=global_limits,
                log_transform=log_transform,
                volumetric=False,  # Processing a single plane, so volumetric=False
                **kwargs,
            )
        return thresh_out

    if log_transform:
        image, conversion_dict = centrosome.threshold.log_transform(image)
    bin_wanted = 0 if assign_middle_to_foreground.casefold() == "foreground" else 1

    thresh_out = None

    if len(image) == 0 or numpy.all(image == numpy.nan):
        thresh_out = numpy.zeros_like(image)
    elif numpy.all(image == image.ravel()[0]):
        thresh_out = numpy.full_like(image, image.ravel()[0])
    # Define the threshold method to be run in each adaptive window
    elif threshold_method.casefold() == "otsu":
        threshold_fn = skimage.filters.threshold_otsu
    elif threshold_method.casefold() == "multiotsu":
        threshold_fn = skimage.filters.threshold_multiotsu
        # If nbins not set in kwargs, use default 128
        kwargs["nbins"] = kwargs["nbins"] if "nbins" in kwargs else 128
    elif threshold_method.casefold() == "minimum_cross_entropy":
        tol = max(numpy.min(numpy.diff(numpy.unique(image))) / 2, 0.5 / 65536)
        kwargs["tolerance"] = tol
        threshold_fn = skimage.filters.threshold_li
    elif threshold_method.casefold() == "robust_background":
        threshold_fn = get_threshold_robust_background
        kwargs["lower_outlier_fraction"] = (
            kwargs["lower_outlier_fraction"]
            if "lower_outlier_fraction" in kwargs
            else 0.05
        )
        kwargs["upper_outlier_fraction"] = (
            kwargs["upper_outlier_fraction"]
            if "upper_outlier_fraction" in kwargs
            else 0.05
        )
        kwargs["averaging_method"] = (
            kwargs["averaging_method"] if "averaging_method" in kwargs else "mean"
        )
        kwargs["variance_method"] = (
            kwargs["variance_method"]
            if "variance_method" in kwargs
            else "standard_deviation"
        )
        kwargs["number_of_deviations"] = (
            kwargs["number_of_deviations"] if "number_of_deviations" in kwargs else 2
        )
    elif threshold_method.casefold() == "sauvola":
        if window_size % 2 == 0:
            window_size += 1
        thresh_out = skimage.filters.threshold_sauvola(image, window_size)
    else:
        raise NotImplementedError(f"Threshold method {threshold_method} not supported.")

    if thresh_out is None:
        image_size = numpy.array(image.shape[:2], dtype=int)
        nblocks = image_size // window_size
        if any(n < 2 for n in nblocks):
            raise ValueError(
                "Adaptive window cannot exceed 50%% of an image dimension.\n"
                "Window of %dpx is too large for a %sx%s image"
                % (window_size, image_size[1], image_size[0])
            )
        #
        # Use a floating point block size to apportion the roundoff
        # roughly equally to each block
        #
        increment = numpy.array(image_size, dtype=float) / numpy.array(
            nblocks, dtype=float
        )
        #
        # Put the answer here
        #
        thresh_out = numpy.zeros(image_size, image.dtype)
        #
        # Loop once per block, computing the "global" threshold within the
        # block.
        #
        block_threshold = numpy.zeros([nblocks[0], nblocks[1]])
        for i in range(nblocks[0]):
            i0 = int(i * increment[0])
            i1 = int((i + 1) * increment[0])
            for j in range(nblocks[1]):
                j0 = int(j * increment[1])
                j1 = int((j + 1) * increment[1])
                block = image[i0:i1, j0:j1]
                block = block[~numpy.logical_not(block)]
                if len(block) == 0:
                    threshold_out = 0.0
                elif numpy.all(block == block[0]):
                    # Don't compute blocks with only 1 value.
                    threshold_out = block[0]
                elif threshold_method == "multiotsu" and len(numpy.unique(block)) < 3:
                    # Region within window has only 2 values.
                    # Can't run 3-class otsu on only 2 values.
                    threshold_out = skimage.filters.threshold_otsu(block)
                else:
                    try:
                        threshold_out = threshold_fn(block, **kwargs)
                    except ValueError:
                        # Drop nbins kwarg when multi-otsu fails. See issue #6324 scikit-image
                        threshold_out = threshold_fn(block)
                if isinstance(threshold_out, numpy.ndarray):
                    # Select correct bin if running multiotsu
                    threshold_out = threshold_out[bin_wanted]
                block_threshold[i, j] = threshold_out
        #
        # Use a cubic spline to blend the thresholds across the image to avoid image artifacts
        #
        spline_order = min(3, numpy.min(nblocks) - 1)
        xStart = int(increment[0] / 2)
        xEnd = int((nblocks[0] - 0.5) * increment[0])
        yStart = int(increment[1] / 2)
        yEnd = int((nblocks[1] - 0.5) * increment[1])
        xtStart = 0.5
        xtEnd = image.shape[0] - 0.5
        ytStart = 0.5
        ytEnd = image.shape[1] - 0.5
        block_x_coords = numpy.linspace(xStart, xEnd, nblocks[0])
        block_y_coords = numpy.linspace(yStart, yEnd, nblocks[1])
        adaptive_interpolation = scipy.interpolate.RectBivariateSpline(
            block_x_coords,
            block_y_coords,
            block_threshold,
            bbox=(xtStart, xtEnd, ytStart, ytEnd),
            kx=spline_order,
            ky=spline_order,
        )
        thresh_out_x_coords = numpy.linspace(
            0.5, int(nblocks[0] * increment[0]) - 0.5, thresh_out.shape[0]
        )
        thresh_out_y_coords = numpy.linspace(
            0.5, int(nblocks[1] * increment[1]) - 0.5, thresh_out.shape[1]
        )
        # Smooth out the "blocky" adaptive threshold
        thresh_out = adaptive_interpolation(thresh_out_x_coords, thresh_out_y_coords)

    # Get global threshold
    global_threshold = get_global_threshold(
        image,
        mask,
        threshold_method,
        threshold_min,
        threshold_max,
        threshold_correction_factor,
        assign_middle_to_foreground,
        log_transform=log_transform,
    )

    if log_transform:
        # Revert the log transformation
        thresh_out = centrosome.threshold.inverse_log_transform(
            thresh_out, conversion_dict
        )
        global_threshold = centrosome.threshold.inverse_log_transform(
            global_threshold, conversion_dict
        )

    # Apply threshold_correction
    thresh_out *= threshold_correction_factor

    t_min = max(threshold_min, global_threshold * global_limits[0])
    t_max = min(threshold_max, global_threshold * global_limits[1])
    thresh_out[thresh_out < t_min] = t_min
    thresh_out[thresh_out > t_max] = t_max
    return thresh_out


def get_global_threshold(
    image,
    mask=None,
    threshold_method="otsu",
    threshold_min=0,
    threshold_max=1,
    threshold_correction_factor=1,
    assign_middle_to_foreground="foreground",
    log_transform=False,
    **kwargs,
):

    if log_transform:
        image, conversion_dict = centrosome.threshold.log_transform(image)

    if mask is not None:
        # Apply mask and discard masked pixels
        image = image[mask]

    # Shortcuts - Check if image array is empty or all pixels are the same value.
    if len(image) == 0:
        threshold = 0.0
    elif numpy.all(image == image.ravel()[0]):
        # All pixels are the same value
        threshold = image.ravel()[0]

    elif threshold_method.casefold() in ("minimum_cross_entropy", "sauvola"):
        tol = max(numpy.min(numpy.diff(numpy.unique(image))) / 2, 0.5 / 65536)
        threshold = skimage.filters.threshold_li(image, tolerance=tol)
    elif threshold_method.casefold() == "robust_background":
        threshold = get_threshold_robust_background(image, **kwargs)
    elif threshold_method.casefold() == "otsu":
        threshold = skimage.filters.threshold_otsu(image)
    elif threshold_method.casefold() == "multiotsu":
        bin_wanted = 0 if assign_middle_to_foreground.casefold() == "foreground" else 1
        kwargs["nbins"] = kwargs["nbins"] if "nbins" in kwargs else 128
        threshold = skimage.filters.threshold_multiotsu(image, **kwargs)
        threshold = threshold[bin_wanted]
    else:
        raise NotImplementedError(f"Threshold method {threshold_method} not supported.")

    if log_transform:
        threshold = centrosome.threshold.inverse_log_transform(
            threshold, conversion_dict
        )

    threshold *= threshold_correction_factor
    threshold = min(max(threshold, threshold_min), threshold_max)
    return threshold


def apply_threshold(image, threshold, mask=None, smoothing=0):
    if mask is None:
        # Create a fake mask if one isn't provided
        mask = numpy.full(image.shape, True)
    if smoothing == 0:
        return (image >= threshold) & mask, 0
    else:
        # Convert from a scale into a sigma. What I've done here
        # is to structure the Gaussian so that 1/2 of the smoothed
        # intensity is contributed from within the smoothing diameter
        # and 1/2 is contributed from outside.
        sigma = smoothing / 0.6744 / 2.0

    blurred_image = centrosome.smooth.smooth_with_function_and_mask(
        image,
        lambda x: scipy.ndimage.gaussian_filter(x, sigma, mode="constant", cval=0),
        mask,
    )
    return (blurred_image >= threshold) & mask, sigma


def overlay_objects(image, labels, opacity=0.3, max_label=None, seed=None, colormap="jet"):
    cmap = matplotlib.cm.ScalarMappable(cmap=matplotlib.cm.get_cmap(colormap))

    colors = cmap.to_rgba(
        numpy.arange(labels.max() if max_label is None else max_label)
    )[:, :3]

    if seed is not None:
        # Resetting the random seed helps keep object label colors consistent in displays
        # where consistency is important, like RelateObjects.
        numpy.random.seed(seed)

    numpy.random.shuffle(colors)

    if labels.ndim == 3:
        overlay = numpy.zeros(labels.shape + (3,), dtype=numpy.float32)

        for index, plane in enumerate(image):
            unique_labels = numpy.unique(labels[index])

            if unique_labels[0] == 0:
                unique_labels = unique_labels[1:]

            overlay[index] = skimage.color.label2rgb(
                labels[index],
                alpha=opacity,
                bg_color=[0, 0, 0],
                bg_label=0,
                colors=colors[unique_labels - 1],
                image=plane,
            )

        return overlay

    return skimage.color.label2rgb(
        labels,
        alpha=opacity,
        bg_color=[0, 0, 0],
        bg_label=0,
        colors=colors,
        image=image,
    )



from typing import Literal

import centrosome.cpmorphology
import numpy
import scipy.ndimage
import skimage.morphology
import mahotas


def shrink_to_point(labels, fill):
    """
    Remove all pixels but one from filled objects.
    If `fill` = False, thin objects with holes to loops.
    """

    if fill:
        labels=centrosome.cpmorphology.fill_labeled_holes(labels)
    return centrosome.cpmorphology.binary_shrink(labels)

def shrink_defined_pixels(labels, fill, iterations):
    """
    Remove pixels around the perimeter of an object unless
    doing so would change the object’s Euler number `iterations` times. 
    Processing stops automatically when there are no more pixels to
    remove.
    """

    if fill:
        labels=centrosome.cpmorphology.fill_labeled_holes(labels)
    return centrosome.cpmorphology.binary_shrink(
                labels, iterations=iterations
            )

def add_dividing_lines(labels):
    """
    Remove pixels from an object that are adjacent to
    another object’s pixels unless doing so would change the object’s
    Euler number
    """

    adjacent_mask = centrosome.cpmorphology.adjacent(labels)

    thinnable_mask = centrosome.cpmorphology.binary_shrink(labels, 1) != 0

    out_labels = labels.copy()

    out_labels[adjacent_mask & ~thinnable_mask] = 0

    return out_labels

def skeletonize(labels):
    """
    Erode each object to its skeleton.
    """
    return centrosome.cpmorphology.skeletonize_labels(labels)

def despur(labels, iterations):
    """
    Remove or reduce the length of spurs in a skeletonized
    image. The algorithm reduces spur size by `iterations` pixels.
    """
    return centrosome.cpmorphology.spur(
                labels, iterations=iterations
            )

def expand(labels, distance):
    """
    Expand labels by a specified distance.
    """

    background = labels == 0

    distances, (i, j) = scipy.ndimage.distance_transform_edt(
        background, return_indices=True
    )

    out_labels = labels.copy()

    mask = background & (distances <= distance)

    out_labels[mask] = labels[i[mask], j[mask]]

    return out_labels

def expand_until_touching(labels):
    """
    Expand objects, assigning every pixel in the
    image to an object. Background pixels are assigned to the nearest
    object.
    """
    distance = numpy.max(labels.shape)
    return expand(labels, distance)

def expand_defined_pixels(labels, iterations):
    """
    Expand each object by adding background pixels
    adjacent to the image `iterations` times. Processing stops 
    automatically if there are no more background pixels.
    """
    return expand(labels, iterations)

def merge_objects(labels_x, labels_y, dimensions):
    """
    Make overlapping objects combine into a single object, taking 
    on the label of the object from the initial set.

    If an object overlaps multiple objects, each pixel of the added 
    object will be assigned to the closest object from the initial 
    set. This is primarily useful when the same objects appear in 
    both sets.
    """
    output = numpy.zeros_like(labels_x)
    labels_y[labels_y > 0] += labels_x.max()
    indices_x = numpy.unique(labels_x)
    indices_x = indices_x[indices_x > 0]
    indices_y = numpy.unique(labels_y)
    indices_y = indices_y[indices_y > 0]
    # Resolve non-conflicting labels first
    undisputed = numpy.logical_xor(labels_x > 0, labels_y > 0)
    undisputed_x = numpy.setdiff1d(indices_x, labels_x[~undisputed])
    mask = numpy.isin(labels_x, undisputed_x)
    output = numpy.where(mask, labels_x, output)
    labels_x[mask] = 0
    undisputed_y = numpy.setdiff1d(indices_y, labels_y[~undisputed])
    mask = numpy.isin(labels_y, undisputed_y)
    output = numpy.where(mask, labels_y, output)
    labels_y[mask] = 0
    to_segment = numpy.logical_or(labels_x > 0, labels_y > 0)
    if dimensions == 2: 
        distances, (i, j) = scipy.ndimage.distance_transform_edt(
            labels_x == 0, return_indices=True
        )
        output[to_segment] = labels_x[i[to_segment], j[to_segment]]
    if dimensions == 3:
        distances, (i, j, v) = scipy.ndimage.distance_transform_edt(
            labels_x == 0, return_indices=True
        )
        output[to_segment] = labels_x[i[to_segment], j[to_segment], v[to_segment]]
    
    return output

def preserve_objects(labels_x, labels_y):
    """
    Preserve the initial object set. Any overlapping regions from 
    the second set will be ignored in favour of the object from 
    the initial set. 
    """
    labels_y[labels_y > 0] += labels_x.max()
    return numpy.where(labels_x > 0, labels_x, labels_y)

def discard_objects(labels_x, labels_y):
    """
    Discard objects that overlap with objects in the initial set
    """
    output = numpy.zeros_like(labels_x)
    indices_x = numpy.unique(labels_x)
    indices_x = indices_x[indices_x > 0]
    indices_y = numpy.unique(labels_y)
    indices_y = indices_y[indices_y > 0]
    # Resolve non-conflicting labels first
    undisputed = numpy.logical_xor(labels_x > 0, labels_y > 0)
    undisputed_x = numpy.setdiff1d(indices_x, labels_x[~undisputed])
    mask = numpy.isin(labels_x, undisputed_x)
    output = numpy.where(mask, labels_x, output)
    labels_x[mask] = 0
    undisputed_y = numpy.setdiff1d(indices_y, labels_y[~undisputed])
    mask = numpy.isin(labels_y, undisputed_y)
    output = numpy.where(mask, labels_y, output)
    labels_y[mask] = 0

    return numpy.where(labels_x > 0, labels_x, output)

def segment_objects(labels_x, labels_y, dimensions):
    """
    Combine object sets and re-draw segmentation for overlapping
    objects.
    """
    output = numpy.zeros_like(labels_x)
    labels_y[labels_y > 0] += labels_x.max()
    indices_x = numpy.unique(labels_x)
    indices_x = indices_x[indices_x > 0]
    indices_y = numpy.unique(labels_y)
    indices_y = indices_y[indices_y > 0]
    # Resolve non-conflicting labels first
    undisputed = numpy.logical_xor(labels_x > 0, labels_y > 0)
    undisputed_x = numpy.setdiff1d(indices_x, labels_x[~undisputed])
    mask = numpy.isin(labels_x, undisputed_x)
    output = numpy.where(mask, labels_x, output)
    labels_x[mask] = 0
    undisputed_y = numpy.setdiff1d(indices_y, labels_y[~undisputed])
    mask = numpy.isin(labels_y, undisputed_y)
    output = numpy.where(mask, labels_y, output)
    labels_y[mask] = 0

    to_segment = numpy.logical_or(labels_x > 0, labels_y > 0)
    disputed = numpy.logical_and(labels_x > 0, labels_y > 0)
    seeds = numpy.add(labels_x, labels_y)
    # Find objects which will be completely removed due to 100% overlap.
    will_be_lost = numpy.setdiff1d(labels_x[disputed], labels_x[~disputed])
    # Check whether this was because an identical object is in both arrays.
    for label in will_be_lost:
        x_mask = labels_x == label
        y_lab = numpy.unique(labels_y[x_mask])
        if not y_lab or len(y_lab) > 1:
            # Labels are not identical
            continue
        else:
            # Get mask of object on y, check if identical to x
            y_mask = labels_y == y_lab[0]
            if numpy.array_equal(x_mask, y_mask):
                # Label is identical
                output[x_mask] = label
                to_segment[x_mask] = False
    seeds[disputed] = 0
    if dimensions == 2:
        distances, (i, j) = scipy.ndimage.distance_transform_edt(
            seeds == 0, return_indices=True
        )
        output[to_segment] = seeds[i[to_segment], j[to_segment]]
    elif dimensions == 3:
        distances, (i, j, v) = scipy.ndimage.distance_transform_edt(
            seeds == 0, return_indices=True
        )
        output[to_segment] = seeds[i[to_segment], j[to_segment], v[to_segment]]

    return output

def watershed(
    input_image: numpy.ndarray,
    mask: numpy.ndarray = None,
    watershed_method: Literal["distance", "markers", "intensity"] = "distance",
    declump_method: Literal["shape", "intensity", "none"] = "shape",
    seed_method: Literal["local", "regional"] = "local",
    intensity_image: numpy.ndarray = None,
    markers_image: numpy.ndarray = None,
    max_seeds: int = -1,
    downsample: int = 1,
    min_distance: int = 1,
    min_intensity: float = 0,
    footprint: int = 8,
    connectivity: int = 1,
    compactness: float = 0.0,
    exclude_border: bool = False,
    watershed_line: bool = False,
    gaussian_sigma: float = 0.0,
    structuring_element: Literal[
        "ball", "cube", "diamond", "disk", "octahedron", "square", "star"
    ] = "disk",
    structuring_element_size: int = 1,
    return_seeds: bool = False,
):
    # Check inputs
    if not numpy.array_equal(input_image, input_image.astype(bool)):
        raise ValueError("Watershed expects a thresholded image as input")
    if (
        watershed_method.casefold() == "intensity" or declump_method.casefold() == "intensity"
    ) and intensity_image is None:
        raise ValueError(f"Intensity-based methods require an intensity image")

    if watershed_method.casefold() == "markers" and markers_image is None:
        raise ValueError("Markers watershed method require a markers image")

    # No declumping, so just label the binary input image
    if declump_method.casefold() == "none":
        if mask is not None:
            input_image[~mask] = 0
        watershed_image = scipy.ndimage.label(input_image)[0]
        if return_seeds:
            return watershed_image, numpy.zeros_like(watershed_image, bool)
        else:
            return watershed_image

    # Create and check structuring element for seed dilation
    strel = getattr(skimage.morphology, structuring_element.casefold())(
        structuring_element_size
    )

    if strel.ndim != input_image.ndim:
        raise ValueError(
            "Structuring element does not match object dimensions: "
            "{} != {}".format(strel.ndim, input_image.ndim)
        )

    if input_image.ndim == 3:
        maxima_footprint = numpy.ones((footprint, footprint, footprint))
    else:
        maxima_footprint = numpy.ones((footprint, footprint))

    # Downsample input image
    if downsample > 1:
        input_shape = input_image.shape
        if input_image.ndim > 2:
            # Only scale x and y
            factors = (1, downsample, downsample)
        else:
            factors = (downsample, downsample)

        input_image = skimage.transform.downscale_local_mean(input_image, factors)
        # Resize optional images
        if intensity_image is not None:
            intensity_image = skimage.transform.downscale_local_mean(
                intensity_image, factors
            )
        if markers_image is not None:
            markers_image = skimage.transform.downscale_local_mean(
                markers_image, factors
            )
        if mask is not None:
            mask = skimage.transform.downscale_local_mean(mask, factors)

    # Only calculate the distance transform if required for shape-based declumping
    # or distance-based seed generation
    if declump_method.casefold() == "shape" or watershed_method.casefold() == "distance":
        smoothed_input_image = skimage.filters.gaussian(
            input_image, sigma=gaussian_sigma
        )
        # Calculate distance transform
        distance = scipy.ndimage.distance_transform_edt(smoothed_input_image)

    # Generate alternative input to the watershed based on declumping
    if declump_method.casefold() == "shape":
        # Invert the distance transform of the input image.
        # The peaks of the distance tranform become the troughs and
        # this image is given as input to watershed
        watershed_input_image = -distance
        # Move to positive realm
        watershed_input_image = watershed_input_image - watershed_input_image.min()
    elif declump_method.casefold() == "intensity":
        # Convert pixel intensity peaks to troughs and
        # use this as the image input in watershed
        watershed_input_image = 1 - intensity_image
    else:
        raise ValueError(f"declump_method {declump_method} is not supported.")

    # Determine image from which to calculate seeds
    if watershed_method.casefold() == "distance":
        seed_image = distance
    elif watershed_method.casefold() == "intensity":
        seed_image = intensity_image
    elif watershed_method.casefold() == "markers":
        # The user has provided their own seeds/markers
        seeds = markers_image
        seeds = skimage.morphology.binary_dilation(seeds, strel)
    else:
        raise NotImplementedError(
            f"watershed method {watershed_method} is not supported"
        )

    if not watershed_method.casefold() == "markers":
        # Generate seeds
        if seed_method.casefold() == "local":
            seed_coords = skimage.feature.peak_local_max(
                seed_image,
                min_distance=min_distance,
                threshold_rel=min_intensity,
                footprint=maxima_footprint,
                num_peaks=max_seeds if max_seeds != -1 else numpy.inf,
                exclude_border=False
            )
            seeds = numpy.zeros(seed_image.shape, dtype=bool)
            seeds[tuple(seed_coords.T)] = True
            seeds = skimage.morphology.binary_dilation(seeds, strel)
            seeds = scipy.ndimage.label(seeds)[0]

        elif seed_method.casefold() == "regional":
            seeds = mahotas.regmax(seed_image, maxima_footprint)
            seeds = skimage.morphology.binary_dilation(seeds, strel)
            seeds = scipy.ndimage.label(seeds)[0]
        else:
            raise NotImplementedError(
                f"seed_method {seed_method} is not supported."
            )

    # Run watershed
    watershed_image = skimage.segmentation.watershed(
        watershed_input_image,
        markers=seeds,
        mask=mask if mask is not None else input_image != 0,
        connectivity=connectivity,
        compactness=compactness,
        watershed_line=watershed_line,
    )

    # Reverse downsampling
    if downsample > 1:
        watershed_image = skimage.transform.resize(
            watershed_image, input_shape, mode="edge", order=0, preserve_range=True
        )
        watershed_image = numpy.rint(watershed_image).astype(numpy.uint16)

    if exclude_border:
        watershed_image = skimage.segmentation.clear_border(watershed_image)

    if return_seeds:
        # Reverse seed downsampling
        if downsample > 1:
            seeds = skimage.transform.resize(
                seeds, input_shape, mode="edge", order=0, preserve_range=True
            )
            seeds = numpy.rint(seeds).astype(numpy.uint16)
        return watershed_image, seeds
    else:
        return watershed_image

def fill_object_holes(labels, diameter, planewise=False):
    array = labels.copy()
    # Calculate radius from diameter
    radius = diameter / 2.0

    # Check if grayscale, RGB or operation is being performed planewise
    if labels.ndim == 2 or labels.shape[-1] in (3, 4) or planewise:
        # 2D circle area will be calculated
        factor = radius ** 2
    else:
        # Calculate the volume of a sphere
        factor = (4.0 / 3.0) * (radius ** 3)

    min_obj_size = numpy.pi * factor

    if planewise and labels.ndim != 2 and labels.shape[-1] not in (3, 4):
        for plane in array:
            for obj in numpy.unique(plane):
                if obj == 0:
                    continue
                filled_mask = skimage.morphology.remove_small_holes(
                    plane == obj, min_obj_size
                )
                plane[filled_mask] = obj
        return array
    else:
        for obj in numpy.unique(array):
            if obj == 0:
                continue
            filled_mask = skimage.morphology.remove_small_holes(
                array == obj, min_obj_size
            )
            array[filled_mask] = obj
    return array

def fill_convex_hulls(labels):
    data = skimage.measure.regionprops(labels)
    output = numpy.zeros_like(labels)
    for prop in data:
        label = prop["label"]
        bbox = prop["bbox"]
        cmask = prop["convex_image"]
        if len(bbox) <= 4:
            output[bbox[0] : bbox[2], bbox[1] : bbox[3]][cmask] = label
        else:
            output[bbox[0] : bbox[3], bbox[1] : bbox[4], bbox[2] : bbox[5]][
                cmask
            ] = label
    return output



from ..functions.image_processing import (
    get_adaptive_threshold,
    get_global_threshold,
    apply_threshold,
)

def threshold(
    image,
    mask=None,
    threshold_scope="global",
    threshold_method="otsu",
    assign_middle_to_foreground="foreground",
    log_transform=False,
    threshold_correction_factor=1,
    threshold_min=0,
    threshold_max=1,
    window_size=50,
    smoothing=0,
    lower_outlier_fraction=0.05,
    upper_outlier_fraction=0.05,
    averaging_method="mean",
    variance_method="standard_deviation",
    number_of_deviations=2,
    volumetric=False,
    automatic=False,
    **kwargs,
):
    """
    Returns three threshold values and a binary image.
    Thresholds returned are:

    Final threshold: Threshold following application of the
    threshold_correction_factor and clipping to min/max threshold

    orig_threshold: The threshold following either adaptive or global
    thresholding strategies, prior to correction

    guide_threshold: Only produced by adaptive threshold, otherwise None.
    This is the global threshold that constrains the adaptive threshold
    within a certain range, as defined by global_limits (default [0.7, 1.5])
    """

    if automatic:
        # Use automatic settings
        smoothing = 1
        log_transform = False
        threshold_scope = "global"
        threshold_method = "minimum_cross_entropy"

    # Only pass robust_background kwargs when selected as the threshold_method
    if threshold_method.casefold() == "robust_background":
        kwargs = {
            "lower_outlier_fraction": lower_outlier_fraction,
            "upper_outlier_fraction": upper_outlier_fraction,
            "averaging_method": averaging_method,
            "variance_method": variance_method,
            "number_of_deviations": number_of_deviations,
        }

    if threshold_scope.casefold() == "adaptive":
        final_threshold = get_adaptive_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            window_size=window_size,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            threshold_correction_factor=threshold_correction_factor,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            volumetric=volumetric,
            **kwargs,
        )
        orig_threshold = get_adaptive_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            window_size=window_size,
            # If automatic=True, do not correct the threshold
            threshold_min=threshold_min if automatic else 0,
            threshold_max=threshold_max if automatic else 1,
            threshold_correction_factor=threshold_correction_factor if automatic else 1,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            volumetric=volumetric,
            **kwargs,
        )

        guide_threshold = get_global_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            threshold_correction_factor=threshold_correction_factor,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            **kwargs,
        )

        binary_image, sigma = apply_threshold(
            image,
            threshold=final_threshold,
            mask=mask,
            smoothing=smoothing,
        )

        return final_threshold, orig_threshold, guide_threshold, binary_image, sigma

    elif threshold_scope.casefold() == "global":
        final_threshold = get_global_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            threshold_correction_factor=threshold_correction_factor,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            **kwargs,
        )
        orig_threshold = get_global_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            # If automatic=True, do not correct the threshold
            threshold_min=threshold_min if automatic else 0,
            threshold_max=threshold_max if automatic else 1,
            threshold_correction_factor=threshold_correction_factor if automatic else 1,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            **kwargs,
        )
        guide_threshold = None
        binary_image, sigma = apply_threshold(
            image,
            threshold=final_threshold,
            mask=mask,
            smoothing=smoothing,
        )
        return final_threshold, orig_threshold, guide_threshold, binary_image, sigma



from enum import Enum
import numpy as np
from numpy.random.mtrand import RandomState
import scipy.sparse
import centrosome.index

class SPARSE_FIELD(Enum):
    label = "label"
    c = "c"
    t = "t"
    z = "z"
    y = "y"
    x = "x"
    
class DENSE_AXIS(Enum):
    label_idx = 0
    c = 1
    t = 2
    z = 3
    y = 4
    x = 5

SPARSE_FIELDS = tuple([mem.value for mem in SPARSE_FIELD])
SPARSE_AXES_FIELDS = SPARSE_FIELDS[1:]
DENSE_AXIS_NAMES = tuple([mem.name for mem in DENSE_AXIS])
DENSE_SHAPE_NAMES = DENSE_AXIS_NAMES[1:]

# ------ Functions for validating segmentation formats ------

def _validate_dense(dense):
    """
    A 'dense' matrix is a 6 dimensional array with axis order:
    (label_idx, c, t, z, y, x)

    When the 'label_idx' dim = 1, it hosts zero or more non-overlapping labels
    When the 'label_idx' dim > 1, each index hosts one or more non-overlapping
    labels (within that index)
    In other words, while labels within an index of 'label_idx' are never
    overlapping, labels between indices of 'label_idx' would overlap
    i.e. 'dense.sum(axis=0)' is invalid, producing innaccurate labels

    A 'dense' matrix is usually paired with an array of indices specifying
    which label values are present in which index of the 'label_idx' dim
    (see 'indices_from_dense' for more details)
    """
    ndim = len(DENSE_AXIS_NAMES)
    assert type(dense) == np.ndarray, "dense must be ndarray"
    assert dense.ndim == ndim, \
    f"dense must be {ndim}-dimensional - f{DENSE_AXIS_NAMES}"

def _validate_dense_shape(dense_shape):
    """
    'dense_shape', as opposed to 'dense.shape', is the shape of the 'dense'
    matrix sans the 'label_idx' axis, i.e.
    (c, t, z, y, z)
    """
    ndim = len(DENSE_SHAPE_NAMES)
    assert (dense_shape is None or
            len(dense_shape) == ndim
    ), f"dense_shape must be length {ndim}, omitting '{DENSE_AXIS.label_idx.name}' dim"

def _validate_labels(labels):
    """
    A 'labels' matrix is another, more constrained, dense representation

    It is strictly 2- or 3-dimensional, of shape: (y, x) or (z, y, x)
    A single 'labels' matrix does not allow for overlapping labels within it

    It is essentially a 'dense' of shape (1, 1, 1, 1, y, x), but squeezed
    such that the ('label_idx', 'c', 't', 'z') axes are removed

    For a 'dense' with shape (2+, 1, 1, 1, y, x), a 'label_set' can be
    constructed (see 'convert_dense_to_label_set' for more details)
    """
    assert type(labels) == np.ndarray, "labels must be ndarray"
    assert (
        labels.ndim == 2 or
        labels.ndim == 3
    ), "labels must be 2- or 3-dimensional"

def _validate_sparse(sparse):
    """
    'sparse' is a sparse representation of labelings
    It's either a numpy recarray, or castable as such via
    'arr.view(np.recarray)'
    where the data types are typed fields who's names are a subset of:
    set('label', 'c', 't', 'z', 'y', 'x')
    and where the data is a 1-dimensional array of tuples, matching the fields

    e.g.
    rec.array([(0, 0, 0, 1), (0, 1, 0, 1), (1, 0, 0, 1), (1, 1, 0, 1),
               (0, 1, 0, 2), (0, 1, 1, 2), (1, 1, 0, 2), (1, 1, 1, 2)],
              dtype=[('z', '<u2'), ('y', '<u2'), ('x', '<u2'), ('label', 'u1')])

    Note that each field may have its own type, the tuple order matches the
    field order, each tuple is unique by at least one element, and there is
    no need (although permissable) to specify a field if it will be the same
    among all the data values (i.e. 'c' and 't' not specified above as they
    would all be '0')

    Note also there is no tuple who's 'label' value is '0' because including
    all '0' values would make 'sparse' equivilent in memory to 'dense',
    and including any '0' values at all would be including a non-label
    (i.e. background)
    """
    assert (
        type(sparse) == np.ndarray or
        type(sparse) == np.recarray
    ), "sparse must be ndarray or recarray"

    assert sparse.ndim == 1, "sparse mut be 1-dimensional"

    axes = sparse.dtype.names

    assert axes is not None, "sparse must have dtype fields"

    axes_set = set(axes)
    full_set = set(SPARSE_FIELDS)

    assert len(axes) == len(axes_set), "duplicate dtype fields in sparse"
    assert axes_set.issubset(full_set), "sparse has unknown dtype fields"

def _validate_ijv(ijv):
    """
    'ijv' is another, more constrained, sparse representation

    It is a 2-dimensional array of triplets, of shape: (num_coords, 3)
    It also allows for overlapping labels

    Unlike 'sparse', it is strictly an ndarray with a single dtype
    for all values, and is strictly in triplet form, where
    [i, j, v] = [y_coord, x_coord, label_value]
    It cannot host values for 'c', 't' or 'z'

    e.g. this 'sparse' record:
    rec.array([(0, 0, 0, 1), (0, 1, 0, 1), (0, 1, 0, 2), (0, 1, 1, 2),],
             dtype=[('z', '<u2'), ('y', '<u2'), ('x', '<u2'), ('label', 'u1')])

    would equate to this 'ijv' matrix:
    array([[0, 0, 1],
           [1, 0, 1],
           [1, 0, 2],
           [1, 1, 2]],
          dtype=uint16)
    """
    assert type(ijv) == np.ndarray, "ijv must be ndarray"
    assert ijv.ndim == 2, "ijv must be 2-dimensional"
    assert ijv.shape[1] == 3, "ijv must have 3 columns"

# ------ Functions converting between segmentation formats ------

def indices_from_dense(dense, validate=True):
    """
    Retrieve indices of a 'dense' matrix

    The return value is a python list of 1-dimensional ndarrays
    whose lengths may or may not be equal (non-homogeneous)

    A given index of the 'indices' list corresponds to the same index in the
    'label_idx' axis of the 'dense' matrix
    An element of the 'indices' list is a 1-dimensional ndarray of labels that
    are present in the 'dense' matrix, at that index

    e.g. the following 'indices' list:
    [array([1, 2, 4], dtype=uint8), array([3, 5], dtype=uint8)]

    specifies that its corresponding 'dense' matrix has a 'label_idx' dim = 2,
    i.e. a shape of: (2, c, t, z, y, x),
    where the labels '1', '2', '4' are present at index '0', and the labels '3'
    and '4' are present at index '1'
    """
    if validate:
        _validate_dense(dense)

    indices = [np.unique(d) for d in dense]
    indices = [idx[1:] if idx[0] == 0 else idx for idx in indices]
    return indices

def dense_shape_from_sparse(sparse, validate=True):
    if validate:
        _validate_sparse(sparse)

    return tuple([
        np.max(sparse[axis]) + 2
        if axis in list(sparse.dtype.fields.keys())
        else 1
        for axis in SPARSE_AXES_FIELDS
    ])

def indices_from_ijv(ijv, validate=True):
    """
    Get the indices for a scipy.ndimage-style function from the 'ijv' formatted
    segmented labels
    """
    if validate:
        _validate_ijv(ijv)

    if len(ijv) == 0:
        return np.zeros(0, np.int32)

    max_label = np.max(ijv[:, 2])

    return np.arange(max_label).astype(np.int32) + 1

def count_from_ijv(ijv, indices=None, validate=True):
    """
    The number of labels in an 'ijv' formatted label matrix
    """
    if validate:
        _validate_ijv(ijv)

    if indices is None:
        indices = indices_from_ijv(ijv, validate=False)

    return len(indices)

def areas_from_ijv(ijv, indices=None, validate=True):
    """
    The area of each object in an 'ijv' formatted label matrix

    Because of the discrete nature of the matrix, this is simply equal to the
    occurrence count of each label in the matrix
    """
    if validate:
        _validate_ijv(ijv)

    if indices is None:
        indices = indices_from_ijv(ijv, validate=False)

    if len(indices) == 0:
        return np.zeros(0, int)

    return np.bincount(ijv[:, 2])[indices]

def downsample_labels(labels, validate=True):
    """
    Convert a 'labels' matrix to the smallest possible integer format
    """
    if validate:
        _validate_labels(labels)

    labels_max = np.max(labels)
    if labels_max < 128:
        return labels.astype(np.int8)
    elif labels_max < 32768:
        return labels.astype(np.int16)
    return labels.astype(np.int32)

def convert_dense_to_label_set(dense, indices=None, validate=True):
    """
    Convert a 'dense' matrix into a list of 2-tuples,
    where the number of tuples corresponds to the 'label_idx' dim of the
    'dense' matrix (see '_validate_dense' for details),
    the tuple's first element is a 'labels' matrix
    (see '_validate_labels' for details),
    and the tuple's second element is the 1-d ndarray of labels in the matrix
    (see 'indices_from_dense' for details)
    """
    if validate:
        _validate_dense(dense)
        assert(
            dense.shape[DENSE_AXIS.c.value] == 1 and 
            dense.shape[DENSE_AXIS.t.value] == 1
        ), f"dense must have shape where f{DENSE_AXIS.c.name} = 1 and f{DENSE_AXIS.t.name} = 1"

    if indices is None:
        indices = indices_from_dense(dense, validate=False)

    label_set_len = dense.shape[DENSE_AXIS.label_idx.value]
    squeezed_dense = dense.squeeze()

    if label_set_len == 1:
        return [(squeezed_dense, indices[0])]
    
    return [(squeezed_dense[i], indices[i]) for i in range(label_set_len)]

def indices_from_labels(labels, validate=True):
    if validate:
        _validate_labels(labels)

    return np.unique(labels[labels != 0])

def cast_labels_to_label_set(labels, validate=True):
    """
    Takes in a 'labels' matrix and casts it into a 1-element 'label_set'
    """
    if validate:
        _validate_labels(labels)

    return [(labels, indices_from_labels(labels, validate=False))]

def convert_labels_to_dense(labels, validate=True):
    """
    Convert a 'labels' matrix (e.g. scipy.ndimage.label) to 'dense' matrix
    """
    if validate:
        _validate_labels(labels)

    typed_labels = downsample_labels(labels, validate=False)

    if labels.ndim == 3:
        expand_axes = (
            DENSE_AXIS.label_idx.value,
            DENSE_AXIS.c.value,
            DENSE_AXIS.t.value
        )
    else:
        expand_axes = (
            DENSE_AXIS.label_idx.value,
            DENSE_AXIS.c.value,
            DENSE_AXIS.t.value,
            DENSE_AXIS.z.value
        )

    return np.expand_dims(typed_labels, axis=expand_axes)

def convert_dense_to_sparse(dense, validate=True):
    if validate:
        _validate_dense(dense)

    full_shape = dense.shape
    label_dim = full_shape[DENSE_AXIS.label_idx.value]
    dense_shape = tuple(
        [full_shape[DENSE_AXIS[n].value] for n in DENSE_SHAPE_NAMES]
    )

    axes_labels = np.array(SPARSE_AXES_FIELDS)
    axes = axes_labels[np.where(np.array(dense_shape) > 1)]

    compact = np.squeeze(dense)
    if label_dim == 1:
        compact = np.expand_dims(compact, axis=0)

    coords = np.where(compact != 0)
    labels = compact[coords]
    # no longer need the labels dim
    coords = coords[1:]

    if np.max(compact.shape) < 2 ** 16:
        coords_dtype = np.uint16
    else:
        coords_dtype = np.uint32

    if len(labels) > 0:
        max_label = np.max(labels)
        if max_label < 2 ** 8:
            labels_dtype = np.uint8
        elif max_label < 2 ** 16:
            labels_dtype = np.uint16
        else:
            labels_dtype = np.uint32
    else:
        labels_dtype = np.uint8

    dtype = [(axis, coords_dtype) for axis in axes]
    dtype.append((SPARSE_FIELD.label.value, labels_dtype))
    sparse = np.core.records.fromarrays(list(coords) + [labels], dtype=dtype)

    return sparse

def convert_ijv_to_sparse(ijv, validate=True):
    if validate:
        _validate_ijv(ijv)

    return np.core.records.fromarrays(
        (ijv[:, 0], ijv[:, 1], ijv[:, 2]),
        [
            (SPARSE_FIELD.y.value, ijv.dtype),
            (SPARSE_FIELD.x.value, ijv.dtype),
            (SPARSE_FIELD.label.value, ijv.dtype)
        ],
    )

def convert_sparse_to_ijv(sparse, validate=True):
    if validate:
        _validate_sparse(sparse)

    return np.column_stack([sparse[axis] for axis in (
        SPARSE_FIELD.y.value, SPARSE_FIELD.x.value, SPARSE_FIELD.label.value)
    ])

def convert_labels_to_ijv(labels, validate=True):
    if validate:
        _validate_labels(labels)

    dense = convert_labels_to_dense(labels, validate=False)
    sparse = convert_dense_to_sparse(dense, validate=False)
    ijv = convert_sparse_to_ijv(sparse, validate=False)

    return ijv

def convert_ijv_to_label_set(ijv, dense_shape=None, validate=True):
    if validate:
        _validate_ijv(ijv)

    sparse = convert_ijv_to_sparse(ijv, validate=False)

    if dense_shape is None:
        dense_shape = dense_shape_from_sparse(sparse)

    dense, indices = convert_sparse_to_dense(
        sparse,
        dense_shape=dense_shape,
        validate=False
    )

    label_set = convert_dense_to_label_set(
        dense,
        indices=indices,
        validate=False
    )

    return label_set

def convert_label_set_to_ijv(label_set, validate=True):
    return np.concatenate(
        [convert_labels_to_ijv(l[0], validate) for l in label_set],
        axis=0
    )

def convert_sparse_to_dense(sparse, dense_shape=None, validate=True):
    """
    Convert 'sparse' representation to 'dense' matrix

    Returns 'dense' matrix and corresponding 'indices'
    """
    if validate:
        _validate_sparse(sparse)
        _validate_dense_shape(dense_shape)

    if len(sparse) == 0:
        if dense_shape is None:
            dense_shape = tuple([1 for _ in range(len(DENSE_SHAPE_NAMES))])

        dense = np.expand_dims(
            np.zeros(dense_shape, np.uint8),
            axis=DENSE_AXIS.label_idx.value
        )

        return dense, indices_from_dense(dense, validate=False)

    if dense_shape is None:
        dense_shape = dense_shape_from_sparse(sparse, validate=False)

    #
    # The code below assigns a "color" to each label so that no
    # two labels have the same color
    #
    positional_columns = []
    available_columns = []
    lexsort_columns = []
    for axis in SPARSE_AXES_FIELDS:
        if axis in list(sparse.dtype.fields.keys()):
            positional_columns.append(sparse[axis])
            available_columns.append(sparse[axis])
            lexsort_columns.insert(0, sparse[axis])
        else:
            positional_columns.append(0)
    labels = sparse[SPARSE_FIELD.label.value]
    lexsort_columns.insert(0, labels)

    sort_order = np.lexsort(lexsort_columns)
    n_labels = np.max(labels)
    #
    # Find the first of a run that's different from the rest
    #
    mask = (
        available_columns[0][sort_order[:-1]]
        != available_columns[0][sort_order[1:]]
    )
    for column in available_columns[1:]:
        mask = mask | (column[sort_order[:-1]] != column[sort_order[1:]])
    breaks = np.hstack(([0], np.where(mask)[0] + 1, [len(labels)]))
    firsts = breaks[:-1]
    counts = breaks[1:] - firsts
    #
    # Eliminate the locations that are singly labeled
    #
    mask = counts > 1
    firsts = firsts[mask]
    counts = counts[mask]
    if len(counts) == 0:
        dense = np.zeros([1] + list(dense_shape), labels.dtype)
        dense[tuple([0] + positional_columns)] = labels
        return dense, indices_from_dense(dense, validate=False)
    #
    # There are n * n-1 pairs for each coordinate (n = # labels)
    # n = 1 -> 0 pairs, n = 2 -> 2 pairs, n = 3 -> 6 pairs
    #
    pairs = centrosome.index.all_pairs(np.max(counts))
    pair_counts = counts * (counts - 1)
    #
    # Create an indexer for the inputs (indexes) and for the outputs
    # (first and second of the pairs)
    #
    # Remember idx points into sort_order which points into labels
    # to get the nth label, grouped into consecutive positions.
    #
    output_indexer = centrosome.index.Indexes(pair_counts)
    #
    # The start of the run of overlaps and the offsets
    #
    run_starts = firsts[output_indexer.rev_idx]
    offs = pairs[output_indexer.idx[0], :]
    first = labels[sort_order[run_starts + offs[:, 0]]]
    second = labels[sort_order[run_starts + offs[:, 1]]]
    #
    # And sort these so that we get consecutive lists for each
    #
    pair_sort_order = np.lexsort((second, first))
    #
    # Eliminate dupes
    #
    to_keep = np.hstack(
        ([True], (first[1:] != first[:-1]) | (second[1:] != second[:-1]))
    )
    to_keep = to_keep & (first != second)
    pair_idx = pair_sort_order[to_keep]
    first = first[pair_idx]
    second = second[pair_idx]
    #
    # Bincount each label so we can find the ones that have the
    # most overlap. See cpmorphology.color_labels and
    # Welsh, "An upper bound for the chromatic number of a graph and
    # its application to timetabling problems", The Computer Journal, 10(1)
    # p 85 (1967)
    #
    overlap_counts = np.bincount(first.astype(np.int32))
    #
    # The index to the i'th label's stuff
    #
    indexes = np.cumsum(overlap_counts) - overlap_counts
    #
    # A vector of a current color per label. All non-overlapping
    # objects are assigned to plane 1
    #
    v_color = np.ones(n_labels + 1, int)
    v_color[0] = 0
    #
    # Clear all overlapping objects
    #
    v_color[np.unique(first)] = 0
    #
    # The processing order is from most overlapping to least
    #
    ol_labels = np.where(overlap_counts > 0)[0]
    processing_order = np.lexsort((ol_labels, overlap_counts[ol_labels]))

    for index in ol_labels[processing_order]:
        neighbors = second[indexes[index] : indexes[index] + overlap_counts[index]]
        colors = np.unique(v_color[neighbors])
        if colors[0] == 0:
            if len(colors) == 1:
                # all unassigned - put self in group 1
                v_color[index] = 1
                continue
            else:
                # otherwise, ignore the unprocessed group and continue
                colors = colors[1:]
        # Match a range against the colors array - the first place
        # they don't match is the first color we can use
        crange = np.arange(1, len(colors) + 1)
        misses = crange[colors != crange]
        if len(misses):
            color = misses[0]
        else:
            max_color = len(colors) + 1
            color = max_color
        v_color[index] = color
    #
    # Create the dense matrix by using the color to address the
    # 5-d hyperplane into which we place each label
    #
    dense = np.zeros([np.max(v_color)] + list(dense_shape), labels.dtype)
    slices = tuple([v_color[labels] - 1] + positional_columns)
    dense[slices] = labels
    indices = [np.where(v_color == i)[0] for i in range(1, dense.shape[0] + 1)]

    return dense, indices

# ------ Functions for operating on segmentation formats ------

def make_rgb_outlines(label_set, colors, random_seed=None, validate=True):
    """
    Assign rgb colors to outlines of labels in 'label_set`

    Make outlines, coloring each object differently to distinguish between
    objects that might overlap.

    'label_set': see 'convert_dense_to_label_set'

    'colors': a N x 3 color map to be used to color the outlines
    where N in dim 0 should match the number of unique labels in the
    `label_set`, and values are R, G, and B values normalized to [0, 1]

    'random_seed' when provided, will seed the RNG for permuting colors
    between 'labels' matrices in the 'label_set'
    """
    if validate:
        assert type(colors) == np.ndarray, "'colors' must be ndarray"
        assert (
            colors.ndim == 2 and
            colors.shape[1] == 3
        ), "'colors' must be of shape (N, 3)"
        indices = [i for _, idxs in label_set for i in idxs]
        # >= because technically you can have superflous colors (but don't)
        assert colors.shape[0] >= len(indices), \
            "axis 1 of 'colors' must be equal to the number of unique labels in 'label_set'"
    #
    # Get planes of non-overlapping objects. The idea here is to use
    # the most similar colors in the color space for objects that
    # don't overlap.
    #
    label_outline_set = [
        (centrosome.outline.outline(label), indexes)
        for label, indexes in label_set
    ]
    rgb_image = np.zeros(list(label_outline_set[0][0].shape) + [3], np.float32)
    #
    # Find out how many unique labels in each
    #
    counts = [np.sum(np.unique(l) != 0) for l, _ in label_outline_set]
    if len(counts) == 1 and counts[0] == 0:
        return rgb_image

    if len(colors) < len(label_outline_set):
        # Have to color 2 planes using the same color!
        # There's some chance that overlapping objects will get
        # the same color. Give me more colors to work with please.
        colors = np.vstack([colors] * (1 + len(label_outline_set) // len(colors)))
    r = RandomState()
    r.seed(random_seed)
    alpha = np.zeros(label_outline_set[0][0].shape, np.float32)
    order = np.lexsort([counts])

    for idx, i in enumerate(order):
        max_available = len(colors) / (len(label_outline_set) - idx)
        ncolors = min(counts[i], max_available)
        my_colors = colors[:ncolors]
        colors = colors[ncolors:]
        my_colors = my_colors[r.permutation(np.arange(ncolors))]
        my_labels, indexes = label_outline_set[i]
        color_idx = np.zeros(np.max(indexes) + 1, int)
        color_idx[indexes] = np.arange(len(indexes)) % ncolors
        rgb_image[my_labels != 0, :] += my_colors[
            color_idx[my_labels[my_labels != 0]], :
        ]
        alpha[my_labels != 0] += 1
    rgb_image[alpha > 0, :] /= alpha[alpha > 0][:, np.newaxis]

    return rgb_image

# needs library tests
def find_label_overlaps(parent_labels, child_labels, validate=True):
    """
    Find per pixel overlap of parent labels and child labels

    'parent_labels' - the parents which contain the children in 'labels' format
    'child_labels' - the children to be mapped to a parent in 'labels' format

    Returns a sparse 'coo_matrix' of overlap between each parent and child.
    Note that the first row and column are empty, as these
    correspond to parent and child labels of 0.
    """
    if validate:
        _validate_labels(parent_labels)
        _validate_labels(child_labels)

    parent_count = np.max(parent_labels)
    child_count = np.max(child_labels)
    #
    # If the labels are different shapes, crop to shared shape.
    #
    common_shape = np.minimum(parent_labels.shape, child_labels.shape)

    if parent_labels.ndim == 3:
        parent_labels = parent_labels[
            0 : common_shape[0], 0 : common_shape[1], 0 : common_shape[2]
        ]
        child_labels = child_labels[
            0 : common_shape[0], 0 : common_shape[1], 0 : common_shape[2]
        ]
    else:
        parent_labels = parent_labels[0 : common_shape[0], 0 : common_shape[1]]
        child_labels = child_labels[0 : common_shape[0], 0 : common_shape[1]]

    #
    # Only look at points that are labeled in parent and child
    #
    not_zero = (parent_labels > 0) & (child_labels > 0)
    not_zero_count = np.sum(not_zero)

    #
    # each row (axis = 0) is a parent
    # each column (axis = 1) is a child
    #
    return scipy.sparse.coo_matrix(
        (
            np.ones((not_zero_count,)),
            (parent_labels[not_zero], child_labels[not_zero]),
        ),
        shape=(parent_count + 1, child_count + 1),
    )

# needs library tests
def find_ijv_overlaps(parent_ijv, child_ijv, validate=True):
    """
    Find per pixel overlap of parent labels and child labels

    'parent_ijv' - the parents which contain the children, in 'ijv' format
    'child_ijv' - the children to be mapped to a parent, in 'ijv' format

    Returns a sparse 'csc_matrix' of overlap between each parent and child.
    Note that the first row and column are empty, as these
    correspond to parent and child labels of 0.
    """
    if validate:
        _validate_ijv(parent_ijv)
        _validate_ijv(child_ijv)

    parent_count = 0 if (parent_ijv.shape[0] == 0) else np.max(parent_ijv[:, 2])
    child_count = 0 if (child_ijv.shape[0] == 0) else np.max(child_ijv[:, 2])

    if parent_count == 0 or child_count == 0:
        return np.zeros((parent_count + 1, child_count + 1), int)

    dim_i = max(np.max(parent_ijv[:, 0]), np.max(child_ijv[:, 0])) + 1
    dim_j = max(np.max(parent_ijv[:, 1]), np.max(child_ijv[:, 1])) + 1
    parent_linear_ij = parent_ijv[:, 0] + dim_i * parent_ijv[:, 1].astype(
        np.uint64
    )
    child_linear_ij = child_ijv[:, 0] + dim_i * child_ijv[:, 1].astype(np.uint64)

    parent_matrix = scipy.sparse.coo_matrix(
        (np.ones((parent_ijv.shape[0],)), (parent_ijv[:, 2], parent_linear_ij)),
        shape=(parent_count + 1, dim_i * dim_j),
    )
    child_matrix = scipy.sparse.coo_matrix(
        (np.ones((child_ijv.shape[0],)), (child_linear_ij, child_ijv[:, 2])),
        shape=(dim_i * dim_j, child_count + 1),
    )
    # I surely do not understand the sparse code.  Converting both
    # arrays to csc gives the best peformance... Why not p.csr and
    # c.csc?
    return parent_matrix.tocsc() * child_matrix.tocsc()

def center_of_labels_mass(labels, validate=True):
    if validate:
        _validate_labels(labels)

    indices = indices_from_labels(labels)
    return np.array(
        scipy.ndimage.center_of_mass(np.ones_like(labels), labels, indices)
    )



import warnings

import numpy
import centrosome

from ..functions.image_processing import (
    enhance_edges_sobel,
    enhance_edges_log,
    enhance_edges_prewitt,
    enhance_edges_canny,
)


def enhanceedges(
    image,
    mask=None,
    method="sobel",
    automatic_threshold=True,
    direction="all",
    automatic_gaussian=True,
    sigma=10,
    manual_threshold=0.2,
    threshold_adjustment_factor=1.0,
    automatic_low_threshold=True,
    low_threshold=0.1,
):
    """EnhanceEdges module

    Parameters
    ----------
    image : numpy.array
        Input image
    mask : numpy.array, optional
        Boolean mask, by default None
    method : str, optional
        Enhance edges algorithm to apply to the input image, by default "sobel"
    direction : str, optional
        Applicable to only the Sobel and Prewitt algorithms, by default "all"
    sigma : int, optional
        Applicable to only the Canny and Laplacian of Gaussian algorithms, by default 10. Only considered if automatic_gaussian is False.
    automatic_threshold : bool, optional
        Applicable only to the Canny algorithm, by default True
    manual_threshold : float, optional
        Applicable only to the Canny algorithm, by default 0.2
    threshold_adjustment_factor : float, optional
        Applicable only to the Canny algorithm, by default 1.0
    automatic_low_threshold : bool, optional
        Applicable only to the Canny algorithm, by default True
    low_threshold : float, optional
        Applicable only to the Canny algorithm, by default 0.1

    Returns
    -------
    numpy.array
        Image with enhanced edges
    """

    if not 0 <= low_threshold <= 1:
        warnings.warn(
            f"""low_threshold value of {low_threshold} is outside
            of the [0-1] CellProfiler default."""
        )

    if mask is None:
        mask = numpy.ones(image.shape, bool)

    if method.casefold() == "sobel":
        output_pixels = enhance_edges_sobel(image, mask, direction)
    elif method.casefold() == "log":
        output_pixels = enhance_edges_log(image, mask, sigma)
    elif method.casefold() == "prewitt":
        output_pixels = enhance_edges_prewitt(image, mask, direction)
    elif method.casefold() == "canny":
        output_pixels = enhance_edges_canny(
            image,
            mask,
            auto_threshold=automatic_threshold,
            auto_low_threshold=automatic_low_threshold,
            sigma=sigma,
            low_threshold=low_threshold,
            manual_threshold=manual_threshold,
            threshold_adjustment_factor=threshold_adjustment_factor,
        )
    elif method.casefold() == "roberts":
        output_pixels = centrosome.filter.roberts(image, mask)
    elif method.casefold() == "kirsch":
        output_pixels = centrosome.kirsch.kirsch(image)
    else:
        raise NotImplementedError(f"{method} edge detection method is not implemented.")

    return output_pixels



import os

import numpy
import skimage

def save_object_image_crops(
    input_image,
    input_objects,
    save_dir,
    file_format="tiff8",
    nested_save=False,
    save_names = {"input_filename": None, "input_objects_name": None},
    volumetric=False
    ):
    """
    For a given input_objects array, save crops for each 
    object of the provided input_image.
    """
    # Build save paths
    if nested_save:
        if not save_names["input_filename"] and not save_names["input_objects_name"]:
            raise ValueError("Must provide a save_names['input_filename'] or save_names['input_objects_name'] for nested save.")
        save_path = os.path.join(
            save_dir, 
            save_names["input_filename"] if save_names["input_filename"] else save_names["input_objects_name"],
            )
    else:
        save_path = save_dir

    if not os.path.exists(save_path):
        os.makedirs(save_path, exist_ok=True)

    unique_labels = numpy.unique(input_objects)

    if unique_labels[0] == 0:
        unique_labels = unique_labels[1:]

    labels = input_objects

    if len(input_image.shape) == len(input_objects.shape) + 1 and not volumetric:
        labels = numpy.repeat(
            labels[:, :, numpy.newaxis], input_image.shape[-1], axis=2
        )

    # Construct filename
    save_filename = f"{save_names['input_filename']+'_' if save_names['input_filename'] else ''}{save_names['input_objects_name']+'_' if save_names['input_objects_name'] else ''}"

    save_filenames = []
    
    for label in unique_labels:
        file_extension = "tiff" if "tiff" in file_format else "png"

        label_save_filename = os.path.join(save_path, save_filename + f"{label}.{file_extension}")
        save_filenames.append(label_save_filename)
        mask_in = labels == label
        properties = skimage.measure.regionprops(
                mask_in.astype(int), intensity_image=input_image
            )
        mask = properties[0].intensity_image
        
        if file_format.casefold() == "png":
            skimage.io.imsave(
                label_save_filename,
                skimage.img_as_ubyte(mask),
                check_contrast=False
            )
        elif file_format.casefold() == "tiff8":
            skimage.io.imsave(
                label_save_filename,
                skimage.img_as_ubyte(mask),
                compression=(8,6),
                check_contrast=False,
            )
        elif file_format.casefold() == "tiff16":
            skimage.io.imsave(
                label_save_filename,
                skimage.img_as_uint(mask),
                compression=(8,6),
                check_contrast=False,
            )
        else:
            raise ValueError(f"{file_format} not in 'png', 'tiff8', or 'tiff16'")
    
    return save_filenames

def save_object_masks(
    input_objects,
    save_dir,
    file_format="tiff8",
    nested_save=False,
    save_names = {"input_filename": None, "input_objects_name": None},
    ):
    """
    For a given object array, save objects as individual masks
    """
    # Build save paths
    if nested_save:
        if not save_names["input_filename"] and not save_names["input_objects_name"]:
            raise ValueError("Must provide a save_names['input_filename'] or save_names['input_objects_name'] for nested save.")
        save_path = os.path.join(
            save_dir, 
            save_names["input_filename"] if save_names["input_filename"] else save_names["input_objects_name"],
            )
    else:
        save_path = save_dir

    if not os.path.exists(save_path):
        os.makedirs(save_path, exist_ok=True)

    unique_labels = numpy.unique(input_objects)

    if unique_labels[0] == 0:
        unique_labels = unique_labels[1:]

    labels = input_objects

    # Construct filename
    save_filename = f"{save_names['input_filename']+'_' if save_names['input_filename'] else ''}{save_names['input_objects_name']+'_' if save_names['input_objects_name'] else ''}"

    filenames = []
    
    for label in unique_labels:
        file_extension = "tiff" if "tiff" in file_format else "png"

        label_save_filename = os.path.join(save_path, save_filename + f"{label}.{file_extension}")

        filenames.append(label_save_filename)

        mask = labels == label
        
        if file_format.casefold() == "png":
                skimage.io.imsave(
                    label_save_filename, 
                    skimage.img_as_ubyte(mask), 
                    check_contrast=False
                )
        elif file_format.casefold() == "tiff8":
            skimage.io.imsave(
                label_save_filename,
                skimage.img_as_ubyte(mask),
                compression=(8, 6),
                check_contrast=False,
            )
        elif file_format.casefold() == "tiff16":
            skimage.io.imsave(
                label_save_filename,
                skimage.img_as_uint(mask),
                compression=(8, 6),
                check_contrast=False,
            )
        else:
            raise ValueError(f"{file_format} not in 'png', 'tiff8', or 'tiff16'")
        
    return filenames


from ..functions.object_processing import add_dividing_lines, despur, expand_defined_pixels, expand_until_touching, shrink_defined_pixels, shrink_to_point, skeletonize

def expand_or_shrink_objects(mode,labels,fill=None,iterations=None):
    if mode == 'expand_defined_pixels':
        return expand_defined_pixels(labels,iterations=iterations)
    elif mode == 'expand_infinite':
        return expand_until_touching(labels)
    elif mode == 'shrink_defined_pixels':
        return shrink_defined_pixels(labels,fill=fill,iterations=iterations)
    elif mode == 'shrink_to_point':
        return shrink_to_point(labels,fill=fill)
    elif mode == 'add_dividing_lines':
        return add_dividing_lines(labels)
    elif mode == 'despur':
        return despur(labels,iterations=iterations)
    elif mode == 'skeletonize':
        return skeletonize(labels)


from typing import Literal

import numpy

from ..functions.object_processing import (
    watershed as library_watershed,
)

# Simple wrapper for the object_procceing watershed function
def watershed(
    input_image: numpy.ndarray,
    mask: numpy.ndarray = None,
    watershed_method: Literal["distance", "intensity", "markers"] = "distance",
    declump_method: Literal["shape", "intensity"] = "shape",
    seed_method: Literal["local", "regional"] = "local",
    intensity_image: numpy.ndarray = None,
    markers_image: numpy.ndarray = None,
    max_seeds: int = -1,
    downsample: int = 1,
    min_distance: int = 1,
    min_intensity: float = 0,
    footprint: int = 8,
    connectivity: int = 1,
    compactness: float = 0.0,
    exclude_border: bool = False,
    watershed_line: bool = False,
    gaussian_sigma: float = 0.0,
    structuring_element: Literal[
        "ball", "cube", "diamond", "disk", "octahedron", "square", "star"
    ] = "disk",
    structuring_element_size: int = 1,
    return_seeds: bool = False,
):
    y_data = library_watershed(
        input_image=input_image,
        mask=mask,
        watershed_method=watershed_method,
        declump_method=declump_method,
        seed_method=seed_method,
        intensity_image=intensity_image,
        markers_image=markers_image,
        max_seeds=max_seeds,
        downsample=downsample,
        min_distance=min_distance,
        min_intensity=min_intensity,
        footprint=footprint,
        connectivity=connectivity,
        compactness=compactness,
        exclude_border=exclude_border,
        watershed_line=watershed_line,
        gaussian_sigma=gaussian_sigma,
        structuring_element=structuring_element,
        structuring_element_size=structuring_element_size,
        return_seeds=return_seeds,
    )
    return y_data
